/****************************************************************************
** Meta object code from reading C++ file 'imageviewer.h'
**
** Created by: The Qt Meta Object Compiler version 67 (Qt 5.15.18)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <memory>
#include "../../../../../src/iv/imageviewer.h"
#include <QtCore/qbytearray.h>
#include <QtCore/qmetatype.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'imageviewer.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 67
#error "This file was generated using the moc from 5.15.18. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

QT_BEGIN_MOC_NAMESPACE
QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
struct qt_meta_stringdata_ImageViewer_t {
    QByteArrayData data[71];
    char stringdata0[971];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_ImageViewer_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_ImageViewer_t qt_meta_stringdata_ImageViewer = {
    {
QT_MOC_LITERAL(0, 0, 11), // "ImageViewer"
QT_MOC_LITERAL(1, 12, 4), // "open"
QT_MOC_LITERAL(2, 17, 0), // ""
QT_MOC_LITERAL(3, 18, 6), // "reload"
QT_MOC_LITERAL(4, 25, 14), // "openRecentFile"
QT_MOC_LITERAL(5, 40, 8), // "closeImg"
QT_MOC_LITERAL(6, 49, 6), // "saveAs"
QT_MOC_LITERAL(7, 56, 12), // "saveWindowAs"
QT_MOC_LITERAL(8, 69, 15), // "saveSelectionAs"
QT_MOC_LITERAL(9, 85, 15), // "moveToNewWindow"
QT_MOC_LITERAL(10, 101, 5), // "print"
QT_MOC_LITERAL(11, 107, 18), // "deleteCurrentImage"
QT_MOC_LITERAL(12, 126, 6), // "zoomIn"
QT_MOC_LITERAL(13, 133, 6), // "smooth"
QT_MOC_LITERAL(14, 140, 7), // "zoomOut"
QT_MOC_LITERAL(15, 148, 12), // "zoomToCursor"
QT_MOC_LITERAL(16, 161, 7), // "newzoom"
QT_MOC_LITERAL(17, 169, 10), // "normalSize"
QT_MOC_LITERAL(18, 180, 16), // "fitImageToWindow"
QT_MOC_LITERAL(19, 197, 16), // "fitWindowToImage"
QT_MOC_LITERAL(20, 214, 6), // "zoomok"
QT_MOC_LITERAL(21, 221, 7), // "minsize"
QT_MOC_LITERAL(22, 229, 16), // "fullScreenToggle"
QT_MOC_LITERAL(23, 246, 5), // "about"
QT_MOC_LITERAL(24, 252, 9), // "prevImage"
QT_MOC_LITERAL(25, 262, 9), // "nextImage"
QT_MOC_LITERAL(26, 272, 11), // "toggleImage"
QT_MOC_LITERAL(27, 284, 18), // "toggleWindowGuides"
QT_MOC_LITERAL(28, 303, 25), // "exposureMinusOneTenthStop"
QT_MOC_LITERAL(29, 329, 24), // "exposureMinusOneHalfStop"
QT_MOC_LITERAL(30, 354, 24), // "exposurePlusOneTenthStop"
QT_MOC_LITERAL(31, 379, 23), // "exposurePlusOneHalfStop"
QT_MOC_LITERAL(32, 403, 10), // "gammaMinus"
QT_MOC_LITERAL(33, 414, 9), // "gammaPlus"
QT_MOC_LITERAL(34, 424, 15), // "viewChannelFull"
QT_MOC_LITERAL(35, 440, 14), // "viewChannelRed"
QT_MOC_LITERAL(36, 455, 16), // "viewChannelGreen"
QT_MOC_LITERAL(37, 472, 15), // "viewChannelBlue"
QT_MOC_LITERAL(38, 488, 16), // "viewChannelAlpha"
QT_MOC_LITERAL(39, 505, 20), // "viewChannelLuminance"
QT_MOC_LITERAL(40, 526, 15), // "viewChannelPrev"
QT_MOC_LITERAL(41, 542, 15), // "viewChannelNext"
QT_MOC_LITERAL(42, 558, 13), // "viewColorRGBA"
QT_MOC_LITERAL(43, 572, 12), // "viewColorRGB"
QT_MOC_LITERAL(44, 585, 12), // "viewColor1Ch"
QT_MOC_LITERAL(45, 598, 16), // "viewColorHeatmap"
QT_MOC_LITERAL(46, 615, 16), // "viewSubimagePrev"
QT_MOC_LITERAL(47, 632, 16), // "viewSubimageNext"
QT_MOC_LITERAL(48, 649, 10), // "sortByName"
QT_MOC_LITERAL(49, 660, 10), // "sortByPath"
QT_MOC_LITERAL(50, 671, 15), // "sortByImageDate"
QT_MOC_LITERAL(51, 687, 14), // "sortByFileDate"
QT_MOC_LITERAL(52, 702, 11), // "sortReverse"
QT_MOC_LITERAL(53, 714, 9), // "slideShow"
QT_MOC_LITERAL(54, 724, 9), // "slideLoop"
QT_MOC_LITERAL(55, 734, 11), // "slideNoLoop"
QT_MOC_LITERAL(56, 746, 20), // "setSlideShowDuration"
QT_MOC_LITERAL(57, 767, 7), // "seconds"
QT_MOC_LITERAL(58, 775, 11), // "slideImages"
QT_MOC_LITERAL(59, 787, 14), // "showInfoWindow"
QT_MOC_LITERAL(60, 802, 19), // "showPixelviewWindow"
QT_MOC_LITERAL(61, 822, 15), // "editPreferences"
QT_MOC_LITERAL(62, 838, 16), // "toggleAreaSample"
QT_MOC_LITERAL(63, 855, 10), // "rotateLeft"
QT_MOC_LITERAL(64, 866, 11), // "rotateRight"
QT_MOC_LITERAL(65, 878, 14), // "flipHorizontal"
QT_MOC_LITERAL(66, 893, 12), // "flipVertical"
QT_MOC_LITERAL(67, 906, 13), // "useOCIOAction"
QT_MOC_LITERAL(68, 920, 7), // "checked"
QT_MOC_LITERAL(69, 928, 20), // "ocioColorSpaceAction"
QT_MOC_LITERAL(70, 949, 21) // "ocioDisplayViewAction"

    },
    "ImageViewer\0open\0\0reload\0openRecentFile\0"
    "closeImg\0saveAs\0saveWindowAs\0"
    "saveSelectionAs\0moveToNewWindow\0print\0"
    "deleteCurrentImage\0zoomIn\0smooth\0"
    "zoomOut\0zoomToCursor\0newzoom\0normalSize\0"
    "fitImageToWindow\0fitWindowToImage\0"
    "zoomok\0minsize\0fullScreenToggle\0about\0"
    "prevImage\0nextImage\0toggleImage\0"
    "toggleWindowGuides\0exposureMinusOneTenthStop\0"
    "exposureMinusOneHalfStop\0"
    "exposurePlusOneTenthStop\0"
    "exposurePlusOneHalfStop\0gammaMinus\0"
    "gammaPlus\0viewChannelFull\0viewChannelRed\0"
    "viewChannelGreen\0viewChannelBlue\0"
    "viewChannelAlpha\0viewChannelLuminance\0"
    "viewChannelPrev\0viewChannelNext\0"
    "viewColorRGBA\0viewColorRGB\0viewColor1Ch\0"
    "viewColorHeatmap\0viewSubimagePrev\0"
    "viewSubimageNext\0sortByName\0sortByPath\0"
    "sortByImageDate\0sortByFileDate\0"
    "sortReverse\0slideShow\0slideLoop\0"
    "slideNoLoop\0setSlideShowDuration\0"
    "seconds\0slideImages\0showInfoWindow\0"
    "showPixelviewWindow\0editPreferences\0"
    "toggleAreaSample\0rotateLeft\0rotateRight\0"
    "flipHorizontal\0flipVertical\0useOCIOAction\0"
    "checked\0ocioColorSpaceAction\0"
    "ocioDisplayViewAction"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_ImageViewer[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
      68,   14, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

 // slots: name, argc, parameters, tag, flags
       1,    0,  354,    2, 0x08 /* Private */,
       3,    0,  355,    2, 0x08 /* Private */,
       4,    0,  356,    2, 0x08 /* Private */,
       5,    0,  357,    2, 0x08 /* Private */,
       6,    0,  358,    2, 0x08 /* Private */,
       7,    0,  359,    2, 0x08 /* Private */,
       8,    0,  360,    2, 0x08 /* Private */,
       9,    0,  361,    2, 0x08 /* Private */,
      10,    0,  362,    2, 0x08 /* Private */,
      11,    0,  363,    2, 0x08 /* Private */,
      12,    1,  364,    2, 0x08 /* Private */,
      12,    0,  367,    2, 0x28 /* Private | MethodCloned */,
      14,    1,  368,    2, 0x08 /* Private */,
      14,    0,  371,    2, 0x28 /* Private | MethodCloned */,
      15,    2,  372,    2, 0x08 /* Private */,
      15,    1,  377,    2, 0x28 /* Private | MethodCloned */,
      17,    0,  380,    2, 0x08 /* Private */,
      18,    0,  381,    2, 0x08 /* Private */,
      19,    2,  382,    2, 0x08 /* Private */,
      19,    1,  387,    2, 0x28 /* Private | MethodCloned */,
      19,    0,  390,    2, 0x28 /* Private | MethodCloned */,
      22,    0,  391,    2, 0x08 /* Private */,
      23,    0,  392,    2, 0x08 /* Private */,
      24,    0,  393,    2, 0x08 /* Private */,
      25,    0,  394,    2, 0x08 /* Private */,
      26,    0,  395,    2, 0x08 /* Private */,
      27,    0,  396,    2, 0x08 /* Private */,
      28,    0,  397,    2, 0x08 /* Private */,
      29,    0,  398,    2, 0x08 /* Private */,
      30,    0,  399,    2, 0x08 /* Private */,
      31,    0,  400,    2, 0x08 /* Private */,
      32,    0,  401,    2, 0x08 /* Private */,
      33,    0,  402,    2, 0x08 /* Private */,
      34,    0,  403,    2, 0x08 /* Private */,
      35,    0,  404,    2, 0x08 /* Private */,
      36,    0,  405,    2, 0x08 /* Private */,
      37,    0,  406,    2, 0x08 /* Private */,
      38,    0,  407,    2, 0x08 /* Private */,
      39,    0,  408,    2, 0x08 /* Private */,
      40,    0,  409,    2, 0x08 /* Private */,
      41,    0,  410,    2, 0x08 /* Private */,
      42,    0,  411,    2, 0x08 /* Private */,
      43,    0,  412,    2, 0x08 /* Private */,
      44,    0,  413,    2, 0x08 /* Private */,
      45,    0,  414,    2, 0x08 /* Private */,
      46,    0,  415,    2, 0x08 /* Private */,
      47,    0,  416,    2, 0x08 /* Private */,
      48,    0,  417,    2, 0x08 /* Private */,
      49,    0,  418,    2, 0x08 /* Private */,
      50,    0,  419,    2, 0x08 /* Private */,
      51,    0,  420,    2, 0x08 /* Private */,
      52,    0,  421,    2, 0x08 /* Private */,
      53,    0,  422,    2, 0x08 /* Private */,
      54,    0,  423,    2, 0x08 /* Private */,
      55,    0,  424,    2, 0x08 /* Private */,
      56,    1,  425,    2, 0x08 /* Private */,
      58,    0,  428,    2, 0x08 /* Private */,
      59,    0,  429,    2, 0x08 /* Private */,
      60,    0,  430,    2, 0x08 /* Private */,
      61,    0,  431,    2, 0x08 /* Private */,
      62,    0,  432,    2, 0x08 /* Private */,
      63,    0,  433,    2, 0x08 /* Private */,
      64,    0,  434,    2, 0x08 /* Private */,
      65,    0,  435,    2, 0x08 /* Private */,
      66,    0,  436,    2, 0x08 /* Private */,
      67,    1,  437,    2, 0x08 /* Private */,
      69,    0,  440,    2, 0x08 /* Private */,
      70,    0,  441,    2, 0x08 /* Private */,

 // slots: parameters
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool,   13,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool,   13,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Float, QMetaType::Bool,   16,   13,
    QMetaType::Void, QMetaType::Float,   16,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool, QMetaType::Bool,   20,   21,
    QMetaType::Void, QMetaType::Bool,   20,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Int,   57,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void,
    QMetaType::Void, QMetaType::Bool,   68,
    QMetaType::Void,
    QMetaType::Void,

       0        // eod
};

void ImageViewer::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    if (_c == QMetaObject::InvokeMetaMethod) {
        auto *_t = static_cast<ImageViewer *>(_o);
        (void)_t;
        switch (_id) {
        case 0: _t->open(); break;
        case 1: _t->reload(); break;
        case 2: _t->openRecentFile(); break;
        case 3: _t->closeImg(); break;
        case 4: _t->saveAs(); break;
        case 5: _t->saveWindowAs(); break;
        case 6: _t->saveSelectionAs(); break;
        case 7: _t->moveToNewWindow(); break;
        case 8: _t->print(); break;
        case 9: _t->deleteCurrentImage(); break;
        case 10: _t->zoomIn((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 11: _t->zoomIn(); break;
        case 12: _t->zoomOut((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 13: _t->zoomOut(); break;
        case 14: _t->zoomToCursor((*reinterpret_cast< float(*)>(_a[1])),(*reinterpret_cast< bool(*)>(_a[2]))); break;
        case 15: _t->zoomToCursor((*reinterpret_cast< float(*)>(_a[1]))); break;
        case 16: _t->normalSize(); break;
        case 17: _t->fitImageToWindow(); break;
        case 18: _t->fitWindowToImage((*reinterpret_cast< bool(*)>(_a[1])),(*reinterpret_cast< bool(*)>(_a[2]))); break;
        case 19: _t->fitWindowToImage((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 20: _t->fitWindowToImage(); break;
        case 21: _t->fullScreenToggle(); break;
        case 22: _t->about(); break;
        case 23: _t->prevImage(); break;
        case 24: _t->nextImage(); break;
        case 25: _t->toggleImage(); break;
        case 26: _t->toggleWindowGuides(); break;
        case 27: _t->exposureMinusOneTenthStop(); break;
        case 28: _t->exposureMinusOneHalfStop(); break;
        case 29: _t->exposurePlusOneTenthStop(); break;
        case 30: _t->exposurePlusOneHalfStop(); break;
        case 31: _t->gammaMinus(); break;
        case 32: _t->gammaPlus(); break;
        case 33: _t->viewChannelFull(); break;
        case 34: _t->viewChannelRed(); break;
        case 35: _t->viewChannelGreen(); break;
        case 36: _t->viewChannelBlue(); break;
        case 37: _t->viewChannelAlpha(); break;
        case 38: _t->viewChannelLuminance(); break;
        case 39: _t->viewChannelPrev(); break;
        case 40: _t->viewChannelNext(); break;
        case 41: _t->viewColorRGBA(); break;
        case 42: _t->viewColorRGB(); break;
        case 43: _t->viewColor1Ch(); break;
        case 44: _t->viewColorHeatmap(); break;
        case 45: _t->viewSubimagePrev(); break;
        case 46: _t->viewSubimageNext(); break;
        case 47: _t->sortByName(); break;
        case 48: _t->sortByPath(); break;
        case 49: _t->sortByImageDate(); break;
        case 50: _t->sortByFileDate(); break;
        case 51: _t->sortReverse(); break;
        case 52: _t->slideShow(); break;
        case 53: _t->slideLoop(); break;
        case 54: _t->slideNoLoop(); break;
        case 55: _t->setSlideShowDuration((*reinterpret_cast< int(*)>(_a[1]))); break;
        case 56: _t->slideImages(); break;
        case 57: _t->showInfoWindow(); break;
        case 58: _t->showPixelviewWindow(); break;
        case 59: _t->editPreferences(); break;
        case 60: _t->toggleAreaSample(); break;
        case 61: _t->rotateLeft(); break;
        case 62: _t->rotateRight(); break;
        case 63: _t->flipHorizontal(); break;
        case 64: _t->flipVertical(); break;
        case 65: _t->useOCIOAction((*reinterpret_cast< bool(*)>(_a[1]))); break;
        case 66: _t->ocioColorSpaceAction(); break;
        case 67: _t->ocioDisplayViewAction(); break;
        default: ;
        }
    }
}

QT_INIT_METAOBJECT const QMetaObject ImageViewer::staticMetaObject = { {
    QMetaObject::SuperData::link<QMainWindow::staticMetaObject>(),
    qt_meta_stringdata_ImageViewer.data,
    qt_meta_data_ImageViewer,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *ImageViewer::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *ImageViewer::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_ImageViewer.stringdata0))
        return static_cast<void*>(this);
    return QMainWindow::qt_metacast(_clname);
}

int ImageViewer::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QMainWindow::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 68)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 68;
    } else if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 68)
            *reinterpret_cast<int*>(_a[0]) = -1;
        _id -= 68;
    }
    return _id;
}
struct qt_meta_stringdata_IvInfoWindow_t {
    QByteArrayData data[1];
    char stringdata0[13];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_IvInfoWindow_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_IvInfoWindow_t qt_meta_stringdata_IvInfoWindow = {
    {
QT_MOC_LITERAL(0, 0, 12) // "IvInfoWindow"

    },
    "IvInfoWindow"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_IvInfoWindow[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

void IvInfoWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    (void)_o;
    (void)_id;
    (void)_c;
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject IvInfoWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QDialog::staticMetaObject>(),
    qt_meta_stringdata_IvInfoWindow.data,
    qt_meta_data_IvInfoWindow,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *IvInfoWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *IvInfoWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_IvInfoWindow.stringdata0))
        return static_cast<void*>(this);
    return QDialog::qt_metacast(_clname);
}

int IvInfoWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDialog::qt_metacall(_c, _id, _a);
    return _id;
}
struct qt_meta_stringdata_IvPreferenceWindow_t {
    QByteArrayData data[1];
    char stringdata0[19];
};
#define QT_MOC_LITERAL(idx, ofs, len) \
    Q_STATIC_BYTE_ARRAY_DATA_HEADER_INITIALIZER_WITH_OFFSET(len, \
    qptrdiff(offsetof(qt_meta_stringdata_IvPreferenceWindow_t, stringdata0) + ofs \
        - idx * sizeof(QByteArrayData)) \
    )
static const qt_meta_stringdata_IvPreferenceWindow_t qt_meta_stringdata_IvPreferenceWindow = {
    {
QT_MOC_LITERAL(0, 0, 18) // "IvPreferenceWindow"

    },
    "IvPreferenceWindow"
};
#undef QT_MOC_LITERAL

static const uint qt_meta_data_IvPreferenceWindow[] = {

 // content:
       8,       // revision
       0,       // classname
       0,    0, // classinfo
       0,    0, // methods
       0,    0, // properties
       0,    0, // enums/sets
       0,    0, // constructors
       0,       // flags
       0,       // signalCount

       0        // eod
};

void IvPreferenceWindow::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    (void)_o;
    (void)_id;
    (void)_c;
    (void)_a;
}

QT_INIT_METAOBJECT const QMetaObject IvPreferenceWindow::staticMetaObject = { {
    QMetaObject::SuperData::link<QDialog::staticMetaObject>(),
    qt_meta_stringdata_IvPreferenceWindow.data,
    qt_meta_data_IvPreferenceWindow,
    qt_static_metacall,
    nullptr,
    nullptr
} };


const QMetaObject *IvPreferenceWindow::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *IvPreferenceWindow::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_meta_stringdata_IvPreferenceWindow.stringdata0))
        return static_cast<void*>(this);
    return QDialog::qt_metacast(_clname);
}

int IvPreferenceWindow::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QDialog::qt_metacall(_c, _id, _a);
    return _id;
}
QT_WARNING_POP
QT_END_MOC_NAMESPACE
